# Copyright 2017 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Grouping dataset transformations."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np

from tensorflow.python.data.ops import dataset_ops
from tensorflow.python.data.util import nest
from tensorflow.python.framework import constant_op
from tensorflow.python.framework import dtypes
from tensorflow.python.framework import ops
from tensorflow.python.framework import tensor_shape
from tensorflow.python.ops import array_ops
from tensorflow.python.ops import check_ops
from tensorflow.python.ops import gen_dataset_ops
from tensorflow.python.ops import math_ops


def group_by_reducer(key_func, reducer):
  """A transformation that groups elements and performs a reduction.

  This transformation maps element of a dataset to a key using `key_func` and
  groups the elements by key. The `reducer` is used to process each group; its
  `init_func` is used to initialize state for each group when it is created, the
  `reduce_func` is used to update the state every time an element is mapped to
  the matching group, and the `finalize_func` is used to map the final state to
  an output value.

  Args:
    key_func: A function mapping a nested structure of tensors
      (having shapes and types defined by `self.output_shapes` and
      `self.output_types`) to a scalar `tf.int64` tensor.
    reducer: An instance of `Reducer`, which captures the reduction logic using
      the `init_func`, `reduce_func`, and `finalize_func` functions.

  Returns:
    A `Dataset` transformation function, which can be passed to
    @{tf.data.Dataset.apply}.
  """

  def _apply_fn(dataset):
    """Function from `Dataset` to `Dataset` that applies the transformation."""
    return _GroupByReducerDataset(dataset, key_func, reducer)

  return _apply_fn


def group_by_window(key_func,
                    reduce_func,
                    window_size=None,
                    window_size_func=None):
  """A transformation that groups windows of elements by key and reduces them.

  This transformation maps each consecutive element in a dataset to a key
  using `key_func` and groups the elements by key. It then applies
  `reduce_func` to at most `window_size_func(key)` elements matching the same
  key. All except the final window for each key will contain
  `window_size_func(key)` elements; the final window may be smaller.

  You may provide either a constant `window_size` or a window size determined by
  the key through `window_size_func`.

  Args:
    key_func: A function mapping a nested structure of tensors
      (having shapes and types defined by `self.output_shapes` and
      `self.output_types`) to a scalar `tf.int64` tensor.
    reduce_func: A function mapping a key and a dataset of up to `window_size`
      consecutive elements matching that key to another dataset.
    window_size: A `tf.int64` scalar `tf.Tensor`, representing the number of
      consecutive elements matching the same key to combine in a single
      batch, which will be passed to `reduce_func`. Mutually exclusive with
      `window_size_func`.
    window_size_func: A function mapping a key to a `tf.int64` scalar
      `tf.Tensor`, representing the number of consecutive elements matching
      the same key to combine in a single batch, which will be passed to
      `reduce_func`. Mutually exclusive with `window_size`.

  Returns:
    A `Dataset` transformation function, which can be passed to
    @{tf.data.Dataset.apply}.

  Raises:
    ValueError: if neither or both of {`window_size`, `window_size_func`} are
      passed.
  """
  if (window_size is not None and window_size_func or
      not (window_size is not None or window_size_func)):
    raise ValueError("Must pass either window_size or window_size_func.")

  if window_size is not None:

    def constant_window_func(unused_key):
      return ops.convert_to_tensor(window_size, dtype=dtypes.int64)

    window_size_func = constant_window_func

  assert window_size_func is not None

  def _apply_fn(dataset):
    """Function from `Dataset` to `Dataset` that applies the transformation."""
    return _GroupByWindowDataset(dataset, key_func, reduce_func,
                                 window_size_func)

  return _apply_fn


def bucket_by_sequence_length(element_length_func,
                              bucket_boundaries,
                              bucket_batch_sizes,
                              padded_shapes=None,
                              padding_values=None,
                              pad_to_bucket_boundary=False):
  """A transformation that buckets elements in a `Dataset` by length.

  Elements of the `Dataset` are grouped together by length and then are padded
  and batched.

  This is useful for sequence tasks in which the elements have variable length.
  Grouping together elements that have similar lengths reduces the total
  fraction of padding in a batch which increases training step efficiency.

  Args:
    element_length_func: function from element in `Dataset` to `tf.int32`,
      determines the length of the element, which will determine the bucket it
      goes into.
    bucket_boundaries: `list<int>`, upper length boundaries of the buckets.
    bucket_batch_sizes: `list<int>`, batch size per bucket. Length should be
      `len(bucket_boundaries) + 1`.
    padded_shapes: Nested structure of `tf.TensorShape` to pass to
      @{tf.data.Dataset.padded_batch}. If not provided, will use
      `dataset.output_shapes`, which will result in variable length dimensions
      being padded out to the maximum length in each batch.
    padding_values: Values to pad with, passed to
      @{tf.data.Dataset.padded_batch}. Defaults to padding with 0.
    pad_to_bucket_boundary: bool, if `False`, will pad dimensions with unknown
      size to maximum length in batch. If `True`, will pad dimensions with
      unknown size to bucket boundary minus 1 (i.e., the maximum length in each
      bucket), and caller must ensure that the source `Dataset` does not contain
      any elements with length longer than `max(bucket_boundaries)`.

  Returns:
    A `Dataset` transformation function, which can be passed to
    @{tf.data.Dataset.apply}.

  Raises:
    ValueError: if `len(bucket_batch_sizes) != len(bucket_boundaries) + 1`.
  """
  with ops.name_scope("bucket_by_seq_length"):
    if len(bucket_batch_sizes) != (len(bucket_boundaries) + 1):
      raise ValueError(
          "len(bucket_batch_sizes) must equal len(bucket_boundaries) + 1")

    batch_sizes = constant_op.constant(bucket_batch_sizes, dtype=dtypes.int64)

    def element_to_bucket_id(*args):
      """Return int64 id of the length bucket for this element."""
      seq_length = element_length_func(*args)

      boundaries = list(bucket_boundaries)
      buckets_min = [np.iinfo(np.int32).min] + boundaries
      buckets_max = boundaries + [np.iinfo(np.int32).max]
      conditions_c = math_ops.logical_and(
          math_ops.less_equal(buckets_min, seq_length),
          math_ops.less(seq_length, buckets_max))
      bucket_id = math_ops.reduce_min(array_ops.where(conditions_c))

      return bucket_id

    def window_size_fn(bucket_id):
      # The window size is set to the batch size for this bucket
      window_size = batch_sizes[bucket_id]
      return window_size

    def make_padded_shapes(shapes, none_filler=None):
      padded = []
      for shape in nest.flatten(shapes):
        shape = tensor_shape.TensorShape(shape)
        shape = [
            none_filler if d.value is None else d
            for d in shape
        ]
        padded.append(shape)
      return nest.pack_sequence_as(shapes, padded)

    def batching_fn(bucket_id, grouped_dataset):
      """Batch elements in dataset."""
      batch_size = batch_sizes[bucket_id]
      none_filler = None
      if pad_to_bucket_boundary:
        err_msg = ("When pad_to_bucket_boundary=True, elements must have "
                   "length < max(bucket_boundaries).")
        check = check_ops.assert_less(
            bucket_id,
            constant_op.constant(len(bucket_batch_sizes) - 1,
                                 dtype=dtypes.int64),
            message=err_msg)
        with ops.control_dependencies([check]):
          boundaries = constant_op.constant(bucket_boundaries,
                                            dtype=dtypes.int64)
          bucket_boundary = boundaries[bucket_id]
          none_filler = bucket_boundary - 1
      shapes = make_padded_shapes(
          padded_shapes or grouped_dataset.output_shapes,
          none_filler=none_filler)
      return grouped_dataset.padded_batch(batch_size, shapes, padding_values)

    def _apply_fn(dataset):
      return dataset.apply(
          group_by_window(element_to_bucket_id, batching_fn,
                          window_size_func=window_size_fn))

    return _apply_fn


def _map_x_dataset(map_func):
  """A transformation that maps `map_func` across its input.

  This transformation is similar to `tf.data.Dataset.map`, but in addition to
  supporting dense and sparse tensor inputs, it also supports dataset inputs.

  Args:
    map_func: A function mapping a nested structure of tensors and/or datasets
      (having shapes and types defined by `self.output_shapes` and
     `self.output_types`) to another nested structure of tensors and/or
     datasets.

  Returns:
    Dataset: A `Dataset`.
  """

  def _apply_fn(dataset):
    """Function from `Dataset` to `Dataset` that applies the transformation."""
    return _MapXDataset(dataset, map_func)

  return _apply_fn


def window_dataset(window_size):
  """A transformation that creates window datasets from the input dataset.

  The resulting datasets will contain `window_size` elements (or
  `N % window_size` for the last dataset if `window_size` does not divide the
  number of input elements `N` evenly).

  Args:
    window_size: A `tf.int64` scalar `tf.Tensor`, representing the number of
      consecutive elements of the input dataset to combine into a window.

  Returns:
    Dataset: A `Dataset`.
  """

  def _apply_fn(dataset):
    return _WindowDataset(dataset, window_size)

  return _apply_fn


class _GroupByReducerDataset(dataset_ops.Dataset):
  """A `Dataset` that groups its input and performs a reduction."""

  def __init__(self, input_dataset, key_func, reducer):
    """See `group_by_reducer()` for details."""
    super(_GroupByReducerDataset, self).__init__()

    self._input_dataset = input_dataset

    self._make_key_func(key_func, input_dataset)
    self._make_init_func(reducer.init_func)
    self._make_reduce_func(reducer.reduce_func, input_dataset)
    self._make_finalize_func(reducer.finalize_func)

  def _make_key_func(self, key_func, input_dataset):
    """Make wrapping Defun for key_func."""
    wrapped_func = dataset_ops.StructuredFunctionWrapper(
        key_func, "tf.contrib.data.group_by_reducer()", input_dataset)
    if not (
        wrapped_func.output_types == dtypes.int64 and
        wrapped_func.output_shapes.is_compatible_with(tensor_shape.scalar())):
      raise ValueError(
          "`key_func` must return a single tf.int64 tensor. "
          "Got type=%s and shape=%s"
          % (wrapped_func.output_types, wrapped_func.output_shapes))
    self._key_func = wrapped_func.function

  def _make_init_func(self, init_func):
    """Make wrapping Defun for init_func."""
    wrapped_func = dataset_ops.StructuredFunctionWrapper(
        init_func, "tf.contrib.data.group_by_reducer()",
        input_classes=ops.Tensor, input_shapes=tensor_shape.scalar(),
        input_types=dtypes.int64)
    self._init_func = wrapped_func.function
    self._state_classes = wrapped_func.output_classes
    self._state_shapes = wrapped_func.output_shapes
    self._state_types = wrapped_func.output_types

  def _make_reduce_func(self, reduce_func, input_dataset):
    """Make wrapping Defun for reduce_func."""

    # Iteratively rerun the reduce function until reaching a fixed point on
    # `self._state_shapes`.
    need_to_rerun = True
    while need_to_rerun:

      wrapped_func = dataset_ops.StructuredFunctionWrapper(
          reduce_func, "tf.contrib.data.group_by_reducer()",
          input_classes=(self._state_classes, input_dataset.output_classes),
          input_shapes=(self._state_shapes, input_dataset.output_shapes),
          input_types=(self._state_types, input_dataset.output_types),
          add_to_graph=False)

      # Extract and validate class information from the returned values.
      for new_state_class, state_class in zip(
          nest.flatten(wrapped_func.output_classes),
          nest.flatten(self._state_classes)):
        if not issubclass(new_state_class, state_class):
          raise TypeError(
              "The element classes for the new state must match the initial "
              "state. Expected %s; got %s." %
              (self._state_classes, wrapped_func.output_classes))

      # Extract and validate type information from the returned values.
      for new_state_type, state_type in zip(
          nest.flatten(wrapped_func.output_types),
          nest.flatten(self._state_types)):
        if new_state_type != state_type:
          raise TypeError(
              "The element types for the new state must match the initial "
              "state. Expected %s; got %s." %
              (self._state_types, wrapped_func.output_types))

      # Extract shape information from the returned values.
      flat_state_shapes = nest.flatten(self._state_shapes)
      flat_new_state_shapes = nest.flatten(wrapped_func.output_shapes)
      weakened_state_shapes = [
          original.most_specific_compatible_shape(new)
          for original, new in zip(flat_state_shapes, flat_new_state_shapes)
      ]

      need_to_rerun = False
      for original_shape, weakened_shape in zip(flat_state_shapes,
                                                weakened_state_shapes):
        if original_shape.ndims is not None and (
            weakened_shape.ndims is None or
            original_shape.as_list() != weakened_shape.as_list()):
          need_to_rerun = True
          break

      if need_to_rerun:
        self._state_shapes = nest.pack_sequence_as(self._state_shapes,
                                                   weakened_state_shapes)

    self._reduce_func = wrapped_func.function
    self._reduce_func.add_to_graph(ops.get_default_graph())

  def _make_finalize_func(self, finalize_func):
    """Make wrapping Defun for finalize_func."""
    wrapped_func = dataset_ops.StructuredFunctionWrapper(
        finalize_func, "tf.contrib.data.group_by_reducer()",
        input_classes=self._state_classes, input_shapes=self._state_shapes,
        input_types=self._state_types)
    self._finalize_func = wrapped_func.function
    self._output_classes = wrapped_func.output_classes
    self._output_shapes = wrapped_func.output_shapes
    self._output_types = wrapped_func.output_types

  @property
  def output_classes(self):
    return self._output_classes

  @property
  def output_shapes(self):
    return self._output_shapes

  @property
  def output_types(self):
    return self._output_types

  def _as_variant_tensor(self):
    return gen_dataset_ops.group_by_reducer_dataset(
        self._input_dataset._as_variant_tensor(),  # pylint: disable=protected-access
        self._key_func.captured_inputs,
        self._init_func.captured_inputs,
        self._reduce_func.captured_inputs,
        self._finalize_func.captured_inputs,
        key_func=self._key_func,
        init_func=self._init_func,
        reduce_func=self._reduce_func,
        finalize_func=self._finalize_func,
        **dataset_ops.flat_structure(self))


class _GroupByWindowDataset(dataset_ops.Dataset):
  """A `Dataset` that groups its input and performs a windowed reduction."""

  def __init__(self, input_dataset, key_func, reduce_func, window_size_func):
    """See `group_by_window()` for details."""
    super(_GroupByWindowDataset, self).__init__()

    self._input_dataset = input_dataset

    self._make_key_func(key_func, input_dataset)
    self._make_reduce_func(reduce_func, input_dataset)
    self._make_window_size_func(window_size_func)

  def _make_window_size_func(self, window_size_func):
    """Make wrapping Defun for window_size_func."""
    def window_size_func_wrapper(key):
      return ops.convert_to_tensor(window_size_func(key), dtype=dtypes.int64)
    wrapped_func = dataset_ops.StructuredFunctionWrapper(
        window_size_func_wrapper, "tf.contrib.data.group_by_window()",
        input_classes=ops.Tensor, input_shapes=tensor_shape.scalar(),
        input_types=dtypes.int64)
    if not (
        wrapped_func.output_types == dtypes.int64 and
        wrapped_func.output_shapes.is_compatible_with(tensor_shape.scalar())):
      raise ValueError(
          "`window_size_func` must return a single tf.int64 scalar tensor.")
    self._window_size_func = wrapped_func.function

  def _make_key_func(self, key_func, input_dataset):
    """Make wrapping Defun for key_func."""
    def key_func_wrapper(*args):
      return ops.convert_to_tensor(key_func(*args), dtype=dtypes.int64)
    wrapped_func = dataset_ops.StructuredFunctionWrapper(
        key_func_wrapper, "tf.contrib.data.group_by_window()", input_dataset)
    if not (
        wrapped_func.output_types == dtypes.int64 and
        wrapped_func.output_shapes.is_compatible_with(tensor_shape.scalar())):
      raise ValueError(
          "`key_func` must return a single tf.int64 scalar tensor.")
    self._key_func = wrapped_func.function

  def _make_reduce_func(self, reduce_func, input_dataset):
    """Make wrapping Defun for reduce_func."""
    nested_dataset = dataset_ops._NestedDatasetComponent(input_dataset)  # pylint: disable=protected-access
    wrapped_func = dataset_ops.StructuredFunctionWrapper(
        reduce_func, "tf.contrib.data.reduce_by_window()",
        input_classes=(ops.Tensor, nested_dataset),
        input_shapes=(tensor_shape.scalar(), nested_dataset),
        input_types=(dtypes.int64, nested_dataset),
        experimental_nested_dataset_support=True)
    if not isinstance(
        wrapped_func.output_classes, dataset_ops._NestedDatasetComponent):  # pylint: disable=protected-access
      raise TypeError("`reduce_func` must return a `Dataset` object.")
    self._output_classes = wrapped_func.output_classes.output_classes
    self._output_types = wrapped_func.output_types.output_types
    self._output_shapes = wrapped_func.output_shapes.output_shapes
    self._reduce_func = wrapped_func.function

  @property
  def output_classes(self):
    return self._output_classes

  @property
  def output_shapes(self):
    return self._output_shapes

  @property
  def output_types(self):
    return self._output_types

  def _as_variant_tensor(self):
    return gen_dataset_ops.group_by_window_dataset(
        self._input_dataset._as_variant_tensor(),  # pylint: disable=protected-access
        self._key_func.captured_inputs,
        self._reduce_func.captured_inputs,
        self._window_size_func.captured_inputs,
        key_func=self._key_func,
        reduce_func=self._reduce_func,
        window_size_func=self._window_size_func,
        **dataset_ops.flat_structure(self))


class Reducer(object):
  """A reducer is used for reducing a set of elements.

  A reducer is represented as a tuple of the three functions:
    1) initialization function: key => initial state
    2) reduce function: (old state, input) => new state
    3) finalization function: state => result
  """

  def __init__(self, init_func, reduce_func, finalize_func):
    self._init_func = init_func
    self._reduce_func = reduce_func
    self._finalize_func = finalize_func

  @property
  def init_func(self):
    return self._init_func

  @property
  def reduce_func(self):
    return self._reduce_func

  @property
  def finalize_func(self):
    return self._finalize_func


class _MapXDataset(dataset_ops.Dataset):
  """A `Dataset` that maps a function over elements in its input."""

  def __init__(self, input_dataset, map_func):
    """See `map_x_dataset()` for details."""
    super(_MapXDataset, self).__init__()
    self._input_dataset = input_dataset

    wrapped_func = dataset_ops.StructuredFunctionWrapper(
        map_func,
        "tf.contrib.data.map_x_dataset()",
        input_dataset,
        experimental_nested_dataset_support=True)
    self._output_classes = wrapped_func.output_classes
    self._output_shapes = wrapped_func.output_shapes
    self._output_types = wrapped_func.output_types
    self._map_func = wrapped_func.function

  def _as_variant_tensor(self):
    input_t = self._input_dataset._as_variant_tensor()  # pylint: disable=protected-access
    return gen_dataset_ops.map_dataset(
        input_t,
        self._map_func.captured_inputs,
        f=self._map_func,
        **dataset_ops.flat_structure(self))

  @property
  def output_classes(self):
    return self._output_classes

  @property
  def output_shapes(self):
    return self._output_shapes

  @property
  def output_types(self):
    return self._output_types


class _WindowDataset(dataset_ops.Dataset):
  """A dataset that creates window datasets from the input elements."""

  def __init__(self, input_dataset, window_size):
    """See `window_dataset()` for more details."""
    super(_WindowDataset, self).__init__()
    self._input_dataset = input_dataset
    self._window_size = ops.convert_to_tensor(
        window_size, dtype=dtypes.int64, name="window_size")
    self._output_classes = nest.pack_sequence_as(
        input_dataset.output_classes,
        [
            dataset_ops._NestedDatasetComponent(  # pylint: disable=protected-access
                output_classes=output_class,
                output_shapes=output_shape,
                output_types=output_type)
            for output_class, output_shape, output_type in zip(
                nest.flatten(input_dataset.output_classes),
                nest.flatten(input_dataset.output_shapes),
                nest.flatten(input_dataset.output_types))
        ])
    self._output_shapes = self._output_classes
    self._output_types = self._output_classes

  def _as_variant_tensor(self):
    return gen_dataset_ops.window_dataset(
        self._input_dataset._as_variant_tensor(),  # pylint: disable=protected-access
        self._window_size,
        **dataset_ops.flat_structure(self))

  @property
  def output_classes(self):
    return self._output_classes

  @property
  def output_shapes(self):
    return self._output_shapes

  @property
  def output_types(self):
    return self._output_types



# Copyright 2017 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Registry for layers and their parameters/variables.

This represents the collection of all layers in the approximate Fisher
information matrix to which a particular FisherBlock may belong. That is, we
might have several layer collections for one TF graph (if we have multiple K-FAC
optimizers being used, for example.)
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# pylint: disable=unused-import,line-too-long,wildcard-import
from tensorflow.contrib.kfac.python.ops.layer_collection import *
from tensorflow.python.util.all_util import remove_undocumented
# pylint: enable=unused-import,line-too-long,wildcard-import

_allowed_symbols = [
    "get_default_layer_collection",
    "set_default_layer_collection",
    "LayerParametersDict",
    "LayerCollection",
    "APPROX_KRONECKER_NAME",
    "APPROX_DIAGONAL_NAME",
    "APPROX_FULL_NAME",
    "VARIABLE_SCOPE",
    "APPROX_KRONECKER_INDEP_NAME",
    "APPROX_KRONECKER_SERIES_1_NAME",
    "APPROX_KRONECKER_SERIES_2_NAME"
]

remove_undocumented(__name__, allowed_exception_list=_allowed_symbols)



# Copyright 2018 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Self-attention generative adversarial with eager execution.

Code for main model.

Reference [Self-Attention Generative Adversarial
Networks](https://arxiv.org/pdf/1805.08318.pdf)
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np
import tensorflow as tf
from tensorflow.contrib.eager.python.examples.sagan import ops
tfe = tf.contrib.eager


class SelfAttentionModule(tf.keras.Model):
  """Self-attention module composed of convolutional layers."""

  def __init__(self,
               attention_features,
               original_features,
               data_format="channels_first"):
    """Initialize the module.

    Args:
      attention_features: Number of filters for the attention computation.
      original_features: Number of filters of the original Tensor.
      data_format: Either 'channels_first' or 'channels_last'
    """
    super(SelfAttentionModule, self).__init__()
    self.data_format = data_format
    # Matrix multiplication implemented as 2D Convolution
    self.f = tf.keras.layers.Conv2D(
        filters=attention_features,
        kernel_size=1,
        strides=(1, 1),
        data_format=data_format)
    self.g = tf.keras.layers.Conv2D(
        filters=attention_features,
        kernel_size=1,
        strides=(1, 1),
        data_format=data_format)
    self.h = tf.keras.layers.Conv2D(
        filters=original_features,
        kernel_size=1,
        strides=(1, 1),
        data_format=data_format)
    self.scale = tfe.Variable(0., trainable=True)

  def call(self, x):
    f = self.f(x)
    g = self.g(x)
    h = self.h(x)

    f_flatten = ops.flatten_hw(f, data_format=self.data_format)
    g_flatten = ops.flatten_hw(g, data_format=self.data_format)
    h_flatten = ops.flatten_hw(h, data_format=self.data_format)

    s = tf.matmul(g_flatten, f_flatten, transpose_b=True)
    b = tf.nn.softmax(s, axis=-1)
    o = tf.matmul(b, h_flatten)
    y = self.scale * tf.reshape(o, tf.shape(x)) + x

    return y

  def compute_output_shape(self, input_shape):
    return input_shape


class SAGAN(tf.contrib.checkpoint.Checkpointable):
  """Self-attention generative adversarial network."""

  def __init__(self, config):
    """Initialize the model.

    Args:
      config: tf.contrib.training.HParams object; specifies hyperparameters
    """
    super(SAGAN, self).__init__()
    self.config = config
    self.generator = self._construct_generator()
    self.discriminator = self._construct_discriminator()

  def _construct_generator(self):
    """Construct generator."""
    # TODO(lxuechen): Add spectral normalization for WGAN
    axis = 1 if self.config.data_format == "channels_first" else 3

    generator = tf.keras.Sequential()
    generator.add(
        tf.keras.layers.InputLayer(input_shape=(self.config.latent_dim,)))
    generator.add(
        tf.keras.layers.Dense(
            units=np.prod(self.config.g_init_shape), activation=tf.nn.relu))

    if self.config.data_format == "channels_first":
      c, h, w = self.config.g_init_shape
    else:
      h, w, c = self.config.g_init_shape

    # Reshape to NHWC/NCHW
    generator.add(
        ops.BroadenHW(h=h, w=w, c=c, data_format=self.config.data_format))

    filters_list = [c // 2**p for p in range(1, self.config.num_upsamples + 1)]
    filters_list[-1] = 3  # Standard RGB images

    for filters in filters_list[:len(filters_list) // 2]:
      generator.add(
          tf.keras.layers.Conv2DTranspose(
              filters=filters,
              kernel_size=4,
              strides=(2, 2),
              use_bias=False,
              padding="SAME",
              data_format=self.config.data_format))
      generator.add(tf.keras.layers.BatchNormalization(axis=axis))
      generator.add(tf.keras.layers.Activation("relu"))

    # pylint: disable=undefined-loop-variable
    generator.add(
        SelfAttentionModule(
            original_features=filters,
            attention_features=filters // 8,
            data_format=self.config.data_format))
    # pylint: enable=undefined-loop-variable

    for filters in filters_list[len(filters_list) // 2:]:
      generator.add(
          tf.keras.layers.Conv2DTranspose(
              filters=filters,
              kernel_size=4,
              strides=(2, 2),
              use_bias=False,
              padding="SAME",
              data_format=self.config.data_format))
      if filters == 3:
        # Assume Image rescaled to [-1, 1]
        generator.add(tf.keras.layers.Activation("tanh"))
      else:
        generator.add(tf.keras.layers.BatchNormalization(axis=axis))
        generator.add(tf.keras.layers.Activation("relu"))

    return generator

  def _construct_discriminator(self):
    """Construct discriminator."""
    # TODO(lxuechen): Add spectral normalization for WGAN
    discriminator = tf.keras.Sequential()
    discriminator.add(
        tf.keras.layers.InputLayer(input_shape=self.config.image_shape))

    filters_list = [
        self.config.d_init_filters * 2**p
        for p in range(self.config.num_upsamples)
    ]

    for filters in filters_list[:(len(filters_list) + 1) // 2]:
      discriminator.add(
          tf.keras.layers.Conv2D(
              filters=filters,
              kernel_size=4,
              strides=(2, 2),
              padding="SAME",
              data_format=self.config.data_format))
      discriminator.add(tf.keras.layers.LeakyReLU(alpha=.1))

    # pylint: disable=undefined-loop-variable
    discriminator.add(
        SelfAttentionModule(
            original_features=filters,
            attention_features=filters // 8,
            data_format=self.config.data_format))
    # pylint: enable=undefined-loop-variable

    for filters in filters_list[(len(filters_list) + 1) // 2:]:
      discriminator.add(
          tf.keras.layers.Conv2D(
              filters=filters,
              kernel_size=4,
              strides=(2, 2),
              padding="SAME",
              data_format=self.config.data_format))
      discriminator.add(tf.keras.layers.LeakyReLU(alpha=.1))

    discriminator.add(tf.keras.layers.Flatten())
    discriminator.add(tf.keras.layers.Dense(units=1))

    return discriminator

  def compute_loss_and_grads(self, real_images, noise, training=True):
    """Compute loss and gradients for both generator and discriminator."""
    # TODO(lxuechen): Add gradient penalty for discriminator
    with tf.GradientTape() as g_tape, tf.GradientTape() as d_tape:
      real_logits = self.discriminator(real_images, training=training)

      fake_images = self.generator.call(noise, training=training)
      fake_logits = self.discriminator.call(fake_images)

      g_loss = self.compute_g_loss(fake_logits)
      d_loss = self.compute_d_loss(fake_logits, real_logits)

    g_grads = g_tape.gradient(g_loss, self.generator.trainable_variables)
    d_grads = d_tape.gradient(d_loss, self.discriminator.trainable_variables)

    return g_loss, d_loss, g_grads, d_grads

  def compute_g_loss(self, fake_logits):
    return -tf.reduce_mean(fake_logits)  # Hinge loss

  def compute_d_loss(self, fake_logits, real_logits):
    # Hinge loss
    real_loss = tf.reduce_mean(tf.nn.relu(1. - real_logits))
    fake_loss = tf.reduce_mean(tf.nn.relu(1. + fake_logits))
    return real_loss + fake_loss



# Copyright 2016 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""SavedModel contrib support.

SavedModel provides a language-neutral format to save machine-learned models
that is recoverable and hermetic. It enables higher-level systems and tools to
produce, consume and transform TensorFlow models.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# pylint: disable=unused-import,wildcard-import,line-too-long
from tensorflow.contrib.saved_model.python.saved_model.signature_def_utils import *
# pylint: enable=unused-import,widcard-import,line-too-long

from tensorflow.python.util.all_util import remove_undocumented

_allowed_symbols = ["get_signature_def_by_key"]

remove_undocumented(__name__, _allowed_symbols)



# Copyright 2017 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Converting AST to code.

Adapted from Tangent.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# TODO(mdan): Use six for compatibility here.
import atexit
import imp
import os
import tempfile

import astor
import gast

from tensorflow.contrib.autograph.pyct import anno
from tensorflow.contrib.autograph.pyct import ast_util
from tensorflow.contrib.autograph.pyct import origin_info
from tensorflow.contrib.autograph.pyct import parser


def _build_source_map(node, code):
  """Return the Python objects represented by given AST.

  Compiling the AST code this way ensures that the source code is readable by
  e.g. `pdb` or `inspect`.

  Args:
    node: An AST node of the original generated code, before the source code is
      generated.
    code: The string representation of the source code for the newly generated
      code.

  Returns:
    Dict[CodeLocation, OriginInfo], a mapping between the user and AutoGraph
    generated code.
  """
  # After we have the final generated code we reparse it to get the final line
  # numbers. Then we walk through the generated and original ASTs in parallel
  # to build the mapping between the user and generated code.
  new_node = parser.parse_str(code)
  origin_info.resolve(new_node, code)
  source_mapping = {}
  for before, after in ast_util.parallel_walk(node, new_node):
    # Need both checks because if origin information is ever copied over to new
    # nodes then we need to rely on the fact that only the original user code
    # has the origin annotation.
    if (anno.hasanno(before, anno.Basic.ORIGIN) and
        anno.hasanno(after, anno.Basic.ORIGIN)):
      source_info = anno.getanno(before, anno.Basic.ORIGIN)
      new_line_number = anno.getanno(after, anno.Basic.ORIGIN).line_number
      source_mapping[new_line_number] = source_info
  return source_mapping


def ast_to_source(node, indentation='  '):
  """Return the source code of given AST."""
  original_node = node
  if isinstance(node, gast.AST):
    node = gast.gast_to_ast(node)
  generator = astor.codegen.SourceGenerator(indentation, False,
                                            astor.string_repr.pretty_string)
  generator.visit(node)
  generator.result.append('\n')
  # In some versions of Python, literals may appear as actual values. This
  # ensures everything is string.
  code = map(str, generator.result)
  code = astor.source_repr.pretty_source(code).lstrip()
  source_mapping = _build_source_map(original_node, code)

  return code, source_mapping


def ast_to_object(node,
                  indentation='  ',
                  source_prefix=None,
                  delete_on_exit=True):
  """Return the Python objects represented by given AST.

  Compiling the AST code this way ensures that the source code is readable by
  e.g. `pdb` or `inspect`.

  Args:
    node: The code to compile, as an AST object.
    indentation: The string to use for indentation.
    source_prefix: Optional string to print as-is into the source file.
    delete_on_exit: Whether to delete the temporary file used for compilation on
      exit.

  Returns:
    A module object containing the compiled source code.
  Raises:
    ValueError: If ag_source_map__ is already in the namespace of the compiled
    node.
  """
  # code_source_mapping does not yet include the offsets from import statements.
  source, code_source_mapping = ast_to_source(node, indentation=indentation)

  with tempfile.NamedTemporaryFile(mode='w', suffix='.py', delete=False) as f:
    # TODO(znado): move into an _offset_source_map() helper function.
    # Need to offset the generated line numbers by the number of import lines.
    if source_prefix:
      num_import_lines = source_prefix.count('\n') + 1
    else:
      num_import_lines = 0
    source_mapping = {}
    for line_number, original_position in code_source_mapping.items():
      source_map_key = origin_info.CodeLocation(
          file_path=f.name, line_number=line_number + num_import_lines)
      source_mapping[source_map_key] = original_position
    module_name = os.path.basename(f.name[:-3])
    if source_prefix:
      f.write(source_prefix)
      f.write('\n')
    f.write(source)
  if delete_on_exit:
    atexit.register(lambda: os.remove(f.name))
  compiled_node = imp.load_source(module_name, f.name)

  # TODO(znado): Clean this up so we don't need to attach it to the namespace.
  # TODO(znado): This does not work for classes because their methods share a
  # namespace.
  # This attaches the source map which is needed for error handling.  Note that
  # api.to_graph copies this source map into an attribute of the function.
  #
  # We need this so the ag_source_map__ variable is available to the call to
  # rewrite_graph_construction_error in the except block inside each function
  # that handles graph construction errors.
  #
  # We cannot get the rewritten function name until it is too late so templating
  # is hard, and this cleanly fixes the
  # issues encountered with nested functions because this is attached to the
  # outermost one.
  source_map_name = 'ag_source_map__'
  if source_map_name in compiled_node.__dict__:
    raise ValueError('cannot convert %s because is has namespace attribute '
                     '"%s", which is reserved for AutoGraph.' %
                     (compiled_node, source_map_name))
  compiled_node.__dict__[source_map_name] = source_mapping

  return compiled_node, source



# Copyright 2016 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Ops and modules related to factorization."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# pylint: disable=wildcard-import
from tensorflow.contrib.factorization.python.ops.clustering_ops import *
from tensorflow.contrib.factorization.python.ops.factorization_ops import *
from tensorflow.contrib.factorization.python.ops.gmm import *
from tensorflow.contrib.factorization.python.ops.gmm_ops import *
from tensorflow.contrib.factorization.python.ops.kmeans import *
from tensorflow.contrib.factorization.python.ops.wals import *
# pylint: enable=wildcard-import

from tensorflow.python.util.all_util import remove_undocumented

_allowed_symbols = [
    'COSINE_DISTANCE',
    'GMM',
    'gmm',
    'GmmAlgorithm',
    'KMeans',
    'KMEANS_PLUS_PLUS_INIT',
    'KMeansClustering',
    'RANDOM_INIT',
    'SQUARED_EUCLIDEAN_DISTANCE',
    'WALSMatrixFactorization',
    'WALSModel',
]

remove_undocumented(__name__, _allowed_symbols)
